/************************************************************************

(C) Copyright 2009, Vadim von Brzeski. vadimv@numberjack.net.

t-mcmc-inf is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

t-mcmc-inf is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with t-mcmc-inf.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************/

#include <gsl/gsl_blas.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_sf_gamma.h>
#include <iostream>
#include <fstream>
#include <string>
#include <gflags/gflags.h>
#include <vector>
#include <math.h>

DEFINE_int32(nAdapt, 2000,  "number of adaptation iterations");
DEFINE_int32(nBurn,  2000,  "number of burn in iterations");
DEFINE_int32(nSamp, 20000, "number of sampling iterations");
DEFINE_string(data, "data.txt", "the pathname to the input file with data, one line per datum");
DEFINE_string(outdir, ".", "the pathname to the directory for output files");
DEFINE_double(A, 1000, "upper bound of uniform prior on degrees of freedom, nu");
DEFINE_double(tau_mu, 1.0, "initial std dev of Normal proposal dist for mu");
DEFINE_double(tau_lambda, 1.0, "initial std dev of Normal proposal dist for lambda");
DEFINE_double(tau_gamma, 1.0, "initial std dev of Normal proposal dist for gamma");
DEFINE_double(acceptRate, 0.44, "ideal MCMC acceptance rate");
DEFINE_double(acceptRateDelta, 0.05, "ideal MCMC acceptance rate tolerance");

void run(int nAdapt, int nBurn, int nSamp, double A, const gsl_vector * Y, double N, gsl_vector * mu_samples, gsl_vector * lambda_samples, gsl_vector * gamma_samples,
	double tau_mu, double tau_lambda, double tau_gamma, gsl_rng * r);
void runOneIter(int i, double A, const gsl_vector * Y, double N, gsl_vector * mu_samples, gsl_vector * lambda_samples, gsl_vector * gamma_samples,
	double tau_mu, double tau_lambda, double tau_gamma, gsl_rng * r, int& accept_mu, int& accept_lambda, int& accept_gamma);
bool readData(const std::string& filename, gsl_vector *& V, int& nRows, double& avg);
double computeSD(const gsl_vector * V, double nRows, double mean);
void writeOutput(const std::string outdir, const gsl_vector * mu_samples, const gsl_vector * lambda_samples, const gsl_vector * gamma_samples, int nBurn, int nSamp);
double log_posterior(double mu, double lambda, double gamma, const gsl_vector * Y, double N);
bool adjustedSigma(double currentRate, double& sigma);


int main(int argc, char** argv)
{

        if (argc < 2)
        {
                std::string usage = "Generates samples from the posterior inference of a t distribution using MCMC.\n\n";
                usage += "Usage : \nt-mcmc-inf -data=<data file> -nAdapt=<no. adaptations iters> -nBurn=<no. burn-in iters> -nSamp=<no. of sampling iters> -outdir=<output-dir-path> -A=<upper bound on nu> -tau_mu=<sd of proposal dist> -tau_lambda=<sd of proposal dist> -tau_gamma=<sd of proposal dist> \n\n";
                usage += "\nt-mcmc-inf --helpshort explains the options in more detail\n\n";
                std::cerr << usage;
                google::SetUsageMessage("");
                google::ParseCommandLineFlags(&argc, &argv, true);
                exit(0);
        }

        google::ParseCommandLineFlags(&argc, &argv, true);

        int nAdapt = FLAGS_nAdapt;
        int nBurn = FLAGS_nBurn;
        int nSamp = FLAGS_nSamp;
        std::string datafile = FLAGS_data;
        std::string outdir = FLAGS_outdir;
	double tau_mu = FLAGS_tau_mu;
	double tau_lambda = FLAGS_tau_lambda;
	double tau_gamma = FLAGS_tau_gamma;
	double A = FLAGS_A;

	if (nAdapt <= 0 || nBurn <= 0 || nSamp <= 0 || tau_mu <= 0 || tau_lambda <= 0 || tau_gamma <= 0 || A <= 2 || (nAdapt >= nBurn+nSamp) )
	{
		std::cerr << "Error in parameters.  nAdapt, nBurn, nSamp, tau_mu, tau_lambda, tau_gamma, A must all be positive, and nAdapt < nBurn+nSamp\n";
		exit(0);
	}

	// allocate vectors to hold samples
        gsl_vector * mu_samples = gsl_vector_alloc(nBurn+nSamp+1);
        gsl_vector * lambda_samples = gsl_vector_alloc(nBurn+nSamp+1);
        gsl_vector * gamma_samples = gsl_vector_alloc(nBurn+nSamp+1);
        gsl_vector_set_zero(mu_samples);
        gsl_vector_set_zero(lambda_samples);
        gsl_vector_set_zero(gamma_samples);

	// read in the data samples Y from a text file and store in vector; calc mean and SD of data set
	gsl_vector * Y = 0;
	int N = 0;
	double meanY = 0;
	if (!readData(datafile, Y, N, meanY))
	{
		std::cerr << "Could not read datafile " << datafile << "\n";
		exit(0);
	}

	double sdY = computeSD(Y, N, meanY);

	// initial starting values
	double mu_0 = meanY;
	double gamma_0 = 0.5 * ( log(2.0) + log(A) );
	double lambda_0 = log(sdY) + 0.5*log(exp(gamma_0 - 2)) - 0.5*gamma_0;	
	gsl_vector_set(mu_samples, 0, mu_0);
	gsl_vector_set(lambda_samples, 0, lambda_0);
	gsl_vector_set(gamma_samples, 0, gamma_0);

	tau_mu = 2.40 * sdY / sqrt(N);
	tau_lambda = 2.40 * sqrt(2) / sqrt(N);
	tau_gamma = 26 / (exp(gamma_0) * sqrt(N));

        // setup random number generator
        const gsl_rng_type * T;
        gsl_rng * r;
        /* create a generator chosen by the environment variable GSL_RNG_TYPE */
        gsl_rng_env_setup();
        T = gsl_rng_default;
        r = gsl_rng_alloc (T);
        srand(time(NULL));

	run(nAdapt, nBurn, nSamp, A, Y, N, mu_samples, lambda_samples, gamma_samples, tau_mu, tau_lambda, tau_gamma, r);
	
	writeOutput(outdir, mu_samples, lambda_samples, gamma_samples, nBurn, nSamp);

	gsl_vector_free(mu_samples);
	gsl_vector_free(lambda_samples);
	gsl_vector_free(gamma_samples);

        return 0;
}
	

double log_posterior(double mu, double lambda, double gamma, const gsl_vector * Y, double N)
{
	double lnp = N * gsl_sf_lngamma((exp(gamma) + 1)/2.0) - (N+1.0)*lambda - (N/2.0 - 1)*gamma - N * gsl_sf_lngamma(exp(gamma)/2.0);
	double s = 0;
	for (size_t i = 0; i < N; i++)
	{
		s = s + log(1 + (gsl_vector_get(Y,i) - mu)*(gsl_vector_get(Y,i) - mu)/exp(gamma+2.0*lambda));
	}
	s = s * (exp(gamma) + 1)/2.0;
	lnp = lnp - s;
	return lnp;
}

void run(int nAdapt, int nBurn, int nSamp, double A, const gsl_vector * Y, double N, gsl_vector * mu_samples, gsl_vector * lambda_samples, gsl_vector * gamma_samples,
	double tau_mu, double tau_lambda, double tau_gamma, gsl_rng * r)
{
	int iter = 1;
	double accepted_mu = 0;
	double accepted_lambda = 0;
	double accepted_gamma = 0;

	int accept_mu = 0;
	int accept_lambda = 0;
	int accept_gamma = 0;

	// Adaptation //////////////////////////////
	bool mu_ok = false;
	bool lambda_ok = false;
	bool gamma_ok = false;

	std::cout << "Adaptation starting ----------------------------\n";
	// while (iter <= nAdapt || !mu_ok || !lambda_ok || !gamma_ok)
	while (!mu_ok || !lambda_ok || !gamma_ok)
	{
		// run one iteration
		runOneIter(iter, A, Y, N, mu_samples, lambda_samples, gamma_samples, tau_mu, tau_lambda, tau_gamma, r, accept_mu, accept_lambda, accept_gamma);
		accepted_mu += accept_mu;
		accepted_lambda += accept_lambda;
		accepted_gamma += accept_gamma;
		// check the acceptance rates every 100 iterations and adjust sd's as necessary
		if (iter % 100 == 0)
		{
			double r_mu = accepted_mu / 100.0;
			double r_lambda = accepted_lambda / 100.0;
			double r_gamma = accepted_gamma / 100;

			// check if the sd's were adjusted; if not, that means that acceptance rate is w/in range
			if (!adjustedSigma(r_mu, tau_mu)) mu_ok = true;
			else mu_ok = false;
			if (!adjustedSigma(r_lambda, tau_lambda)) lambda_ok = true;
			else lambda_ok = false;
			if (!adjustedSigma(r_gamma, tau_gamma)) gamma_ok = true;
			else gamma_ok = false;

			accepted_mu = 0;
			accepted_lambda = 0;
			accepted_gamma = 0;

			std::cout << "Iterations : " << iter << "\n";
			std::cout << "\tmu_rate = " << r_mu << ".\tlambda_rate = " << r_lambda << ".\tgamma_rate = " << r_gamma << ".\n";
			std::cout << "\ttau_mu = " << tau_mu << ".\ttau_lambda = " << tau_lambda << ".\ttau_gamma = " << tau_gamma << ".\n";
		}
		iter++;
	}
	std::cout << "Adaptation finished --------------------------------- \n";

	// Burn-in and Sampling /////////////////////////////////////

	double last_mu = gsl_vector_get(mu_samples, iter-1);
	double last_lambda = gsl_vector_get(lambda_samples, iter-1);
	double last_gamma = gsl_vector_get(gamma_samples, iter-1);
        gsl_vector_set_zero(mu_samples);
        gsl_vector_set_zero(lambda_samples);
        gsl_vector_set_zero(gamma_samples);
	gsl_vector_set(mu_samples, 0, last_mu);
	gsl_vector_set(lambda_samples, 0, last_lambda);
	gsl_vector_set(gamma_samples, 0, last_gamma);
	iter = 1;
	while (iter <= nBurn + nSamp)
	{
		// run one iteration
		runOneIter(iter, A, Y, N, mu_samples, lambda_samples, gamma_samples, tau_mu, tau_lambda, tau_gamma, r, accept_mu, accept_lambda, accept_gamma);
		if (iter % 100 == 0)
		{
			if (iter <= nBurn) std::cout << "Burn-in : Iterations : " << iter << "\n";
			else std::cout << "Monitoring : Iterations : " << iter << "\n";
		}
		iter++;
	}
	std::cout << "Done !\n";
}


void runOneIter(int i, double A, const gsl_vector * Y, double N, gsl_vector * mu_samples, gsl_vector * lambda_samples, gsl_vector * gamma_samples,
	double tau_mu, double tau_lambda, double tau_gamma, gsl_rng * r, int& accept_mu, int& accept_lambda, int& accept_gamma)
{
	accept_mu = 0;
	accept_lambda = 0;
	accept_gamma = 0;

	double mu_prev = gsl_vector_get(mu_samples, i - 1);
	double lambda_prev = gsl_vector_get(lambda_samples, i - 1);
	double gamma_prev = gsl_vector_get(gamma_samples, i - 1);
	double lp_prev = log_posterior(mu_prev, lambda_prev, gamma_prev, Y, N);

	/////////////////////////////// mu ////////////////////////////////////

	// draw from proposal dist	
        gsl_rng_set (r, rand());
	double mu_star = gsl_ran_gaussian (r, tau_mu) + mu_prev;
	// calculate log posterior
	double lp_star = log_posterior(mu_star, lambda_prev, gamma_prev, Y, N);
	// uphill move...
	if (lp_star > lp_prev)
	{
		// accept mu draw
		accept_mu = 1;
		gsl_vector_set(mu_samples, i, mu_star);
		// set latest value of mu to the draw
		mu_prev = mu_star;
		// calc log posterior needed downstream
		lp_prev = log_posterior(mu_prev, lambda_prev, gamma_prev, Y, N);
	}
	// downhill move...
	else
	{
		// accept w/ prob exp(lp* - lp0)
        	gsl_rng_set (r, rand());
        	int draw = gsl_ran_bernoulli (r, exp(lp_star - lp_prev));
        	if (draw == 1)
		{
			// accept mu draw
			accept_mu = 1;
			gsl_vector_set(mu_samples, i, mu_star);
			// set latest value of mu to the draw
			mu_prev = mu_star;
			// calc log posterior needed downstream
			lp_prev = log_posterior(mu_prev, lambda_prev, gamma_prev, Y, N);
		}
		else
		{
			// reject
			gsl_vector_set(mu_samples, i, mu_prev);
		}
	}

	//////////////////////////// lambda ////////////////////////////////////

	// draw from proposal dist	
        gsl_rng_set (r, rand());
	double lambda_star = gsl_ran_gaussian (r, tau_lambda) + lambda_prev;
	// calculate log posterior
	lp_star = log_posterior(mu_prev, lambda_star, gamma_prev, Y, N);
	// uphill move...
	if (lp_star > lp_prev)
	{
		// accept lambda draw
		accept_lambda = 1;
		gsl_vector_set(lambda_samples, i, lambda_star);
		// set the latest value of lambda to the draw
		lambda_prev = lambda_star;		
		// calc log posterior needed downstream
		lp_prev = log_posterior(mu_prev, lambda_prev, gamma_prev, Y, N);
	}
	// downhill move...
	else
	{
		// accept w/ prob exp(lp* - lp0)
        	gsl_rng_set (r, rand());
        	int draw = gsl_ran_bernoulli (r, exp(lp_star - lp_prev));
        	if (draw == 1)
		{
			// accept lambda draw
			accept_lambda = 1;
			gsl_vector_set(lambda_samples, i, lambda_star);
			// set the latest value of lambda to the draw
			lambda_prev = lambda_star;		
			// calc log posterior needed downstream
			lp_prev = log_posterior(mu_prev, lambda_prev, gamma_prev, Y, N);
		}
		else
		{
			gsl_vector_set(lambda_samples, i, lambda_prev);
		}
	}

	//////////////////////////// gamma ////////////////////////////////////

	// draw from proposal dist - need to respect boundaries
	bool ok = false;
	double gamma_star = 0;
	while (!ok)
	{
        	gsl_rng_set (r, rand());
		gamma_star = gsl_ran_gaussian (r, tau_gamma) + gamma_prev;
		if (gamma_star < log(A) && gamma_star > log(2.0)) ok = true;
	}
	// calculate log posterior
	lp_star = log_posterior(mu_prev, lambda_prev, gamma_star, Y, N);
	if (lp_star > lp_prev)
	{
		// accept gamma draw
		accept_gamma = 1;
		gsl_vector_set(gamma_samples, i, gamma_star);
		// set the latest value of gamma to the draw
		gamma_prev = gamma_star;		
	}
	// downhill move...
	else
	{
		// accept w/ prob exp(lp* - lp0)
        	gsl_rng_set (r, rand());
        	int draw = gsl_ran_bernoulli (r, exp(lp_star - lp_prev));
        	if (draw == 1)
		{
			// accept gamma draw
			accept_gamma;
			gsl_vector_set(gamma_samples, i, gamma_star);
			// set the latest value of gamma to the draw
			gamma_prev = gamma_star;		
		}
		else
		{
			gsl_vector_set(gamma_samples, i, gamma_prev);
		}
	}
}


bool adjustedSigma(double currentRate, double& sigma)
{
	if (currentRate >= FLAGS_acceptRate + FLAGS_acceptRateDelta)
	{
		sigma = sigma * (2 - (1 - currentRate)/(1 - FLAGS_acceptRate));
		return true;
	}
	else if (currentRate <= FLAGS_acceptRate - FLAGS_acceptRateDelta)
	{
		sigma = sigma / (2 - currentRate/FLAGS_acceptRate);
		return true;
	}	
	return false;
}


bool readData(const std::string& filename, gsl_vector *& V, int& nRows, double& avg)
{
        std::ifstream f;
        f.open(filename.c_str(), std::ios::in);
        if (!f.good()) return false;

        std::vector<double> values;
        std::string line;
        while (!f.eof())
        {
                getline(f,line);
                if (line.size() == 0) break;
                values.push_back(atof(line.c_str()));
        }
        nRows = values.size();

        avg = 0;
        V = gsl_vector_alloc(nRows);
        gsl_vector_set_zero(V);
        for (size_t i = 0; i < values.size(); i++)
        {
                gsl_vector_set(V, i, values.at(i));
                avg += values.at(i);
        }
        avg = avg / (double)nRows;

        return true;
}

void writeOutput(const std::string outdir, const gsl_vector * mu_samples, const gsl_vector * lambda_samples, const gsl_vector * gamma_samples, int nBurn, int nSamp)
{
        std::string filename = outdir + "/samples.txt";
        std::ofstream outX(filename.c_str());
        outX << "mu\tlambda\tsigma\tgamma\tnu\n";
        for (size_t i = nBurn+1; i <= nBurn+nSamp; i++)
        {
        	outX << gsl_vector_get(mu_samples, i) << "\t" << gsl_vector_get(lambda_samples, i) << "\t" << exp(gsl_vector_get(lambda_samples, i)) << "\t" << gsl_vector_get(gamma_samples, i) << "\t" << exp(gsl_vector_get(gamma_samples, i)) << "\n";
        }
        outX.close();
	std::cout << "Wrote " << filename << "\n";
}



double computeSD(const gsl_vector * V, double nRows, double mean)
{
        double sigma = 0;
        for (size_t i = 0; i < nRows; i++)
        {
                double v = gsl_vector_get(V,i);
                sigma += (v-mean) * (v-mean);
        }
        sigma = sigma / (nRows - 1);
        return sigma;
}


	

